
/*============================================================================
  bandwidth 0.11, a benchmark to estimate memory transfer bandwidth.
  Copyright (C) 2005, by Zack T Smith.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  The author may be reached at fbui@comcast.net.
 *===========================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <linux/fb.h>

#define VERSION "0.11"

#define MAINCOUNT 2*1024*1024
#define MAINLOOPS_R 10000
#define MAINLOOPS_W 2500

#define L1CACHECOUNT 512
#define L1CACHELOOPS_R 50*1024*1024
#define L1CACHELOOPS_W 50*1024*1024

#define L2CACHECOUNT 32768
#define L2CACHELOOPS_R 2000*1024
#define L2CACHELOOPS_W 2000*1024

#define FBLOOPS_R 10000
#define FBLOOPS_W 50000
#define FBCOUNT 131072


void
main_seq_write ()
{
	volatile unsigned long *chunk = (unsigned long*) malloc (4*MAINCOUNT);
	volatile unsigned long * v;
	unsigned long register counter = 0;
	register int loops;
	double rate;
	time_t t, t0;

	//-------------------------------------------------
	printf ("Main memory sequential write ");
	fflush (stdout);

	counter = MAINCOUNT;
	loops = MAINLOOPS_W;
	v = chunk;

	t0 = time(NULL);
	while (loops--) {
		while (counter--)
			*v++ = 0xDeadBeef;

		v = chunk;
		counter = MAINCOUNT;
	}
	t = time(NULL);

	rate = MAINCOUNT;
	rate *= MAINLOOPS_W;
	rate /= (t-t0);
	rate /= 1000000;

	printf ("millions/sec = %g\n", rate);
	free((void*)chunk);
}

void
main_seq_read ()
{
	volatile unsigned long *chunk = (unsigned long*) malloc (4*MAINCOUNT);
	volatile unsigned long * v;
	unsigned long register counter = 0;
	register int loops;
	double rate;
	time_t t, t0;
	unsigned long chunk2[L1CACHECOUNT];
	register unsigned long foo;

	//-------------------------------------------------
	printf ("Main memory sequential read ");
	fflush (stdout);

	loops = MAINLOOPS_R;
	counter = MAINCOUNT;
	v = chunk;

	t0 = time(NULL);
	while (loops--) {
		while (counter--)
			foo = *v++;

		v = chunk;
		counter = MAINCOUNT;
	}
	t = time(NULL);

	rate = MAINCOUNT;
	rate *= MAINLOOPS_R;
	rate /= (t-t0);
	rate /= 1000000;

	printf ("millions/sec = %g\n", rate);
	free((void*)chunk);
}


void
l1_seq_read ()
{
	unsigned long register counter = 0;
	unsigned long register *ptr;
	register int loops;
	double rate;
	time_t t, t0;
	unsigned long chunk2[L1CACHECOUNT];
	register unsigned long foo;

	//-------------------------------------------------

	ptr = chunk2;
	loops = L1CACHELOOPS_R;
	counter = L1CACHECOUNT;

	printf ("L1 cache sequential read ");
	fflush (stdout);

	t0 = time(NULL);
	while (loops--) {
		while (counter--)
			foo = *ptr++;

		ptr = chunk2;
		counter = L1CACHECOUNT;
	}
	t = time(NULL);

	rate = L1CACHECOUNT;
	rate *= L1CACHELOOPS_R;
	rate /= (t-t0);
	rate /= 1000000;

	printf ("millions/sec = %g\n", rate);
}


void
l1_seq_write ()
{
	unsigned long register counter = 0;
	unsigned long register *ptr;
	register int loops;
	double rate;
	time_t t, t0;
	unsigned long chunk2[L1CACHECOUNT];

	//-------------------------------------------------

	ptr = chunk2;
	loops = L1CACHELOOPS_W;
	counter = L1CACHECOUNT;

	printf ("L1 cache sequential write ");
	fflush (stdout);

	t0 = time(NULL);
	while (loops--) {
		while (counter--)
			*ptr++ = 0xDeadBeef;

		ptr = chunk2;
		counter = L1CACHECOUNT;
	}
	t = time(NULL);

	rate = L1CACHECOUNT;
	rate *= L1CACHELOOPS_W;
	rate /= (t-t0);
	rate /= 1000000;

	printf ("millions/sec = %g\n", rate);
}

void
l2_seq_read ()
{
	unsigned long register counter = 0;
	unsigned long register *ptr;
	register int loops;
	double rate;
	time_t t, t0;
	unsigned long chunk3[L2CACHECOUNT];
	register unsigned long foo;

	//-------------------------------------------------

	ptr = chunk3;
	loops = L2CACHELOOPS_R;
	counter = L2CACHECOUNT;

	printf ("L2 cache sequential read ");
	fflush (stdout);

	t0 = time(NULL);
	while (loops--) {
		while (counter--)
			foo = *ptr++;

		ptr = chunk3;
		counter = L2CACHECOUNT;
	}
	t = time(NULL);

	rate = L2CACHECOUNT;
	rate *= L2CACHELOOPS_R;
	rate /= (t-t0);
	rate /= 1000000;

	printf ("millions/sec = %g\n", rate);
}


void
l2_seq_write ()
{
	unsigned long register counter = 0;
	unsigned long register *ptr;
	register int loops;
	double rate;
	time_t t, t0;
	unsigned long chunk3[L2CACHECOUNT];

	//-------------------------------------------------

	ptr = chunk3;
	loops = L2CACHELOOPS_W;
	counter = L2CACHECOUNT;

	printf ("L2 cache sequential write ");
	fflush (stdout);

	t0 = time(NULL);
	while (loops--) {
		while (counter--)
			*ptr++ = 0xDeadBeef;

		ptr = chunk3;
		counter = L2CACHECOUNT;
	}
	t = time(NULL);

	rate = L2CACHECOUNT;
	rate *= L2CACHELOOPS_W;
	rate /= (t-t0);
	rate /= 1000000;

	printf ("millions/sec = %g\n", rate);
}


void
fb_readwrite ()
{
	volatile unsigned long * v;
	unsigned long register counter = 0;
	register int loops;
	double rate;
	time_t t, t0;
	static struct fb_fix_screeninfo fi;
	static struct fb_var_screeninfo vi;
	unsigned long *fb = NULL;
	unsigned long fblen = 0;
	int fd;
	register unsigned long foo;

	//-------------------------------------------------

	fd = open ("/dev/fb0", O_RDWR);
	if (fd < 0) 
		fd = open ("/dev/fb/0", O_RDWR);
	if (fd < 0) {
		perror ("open");
		printf ("Can't open framebuffer.\n");
		exit(0);
	}

	if (ioctl (fd, FBIOGET_FSCREENINFO, &fi))
	{
		perror("ioctl");
		exit(1);
	}
	else
	if (ioctl (fd, FBIOGET_VSCREENINFO, &vi))
	{
		perror("ioctl");
		exit(1);
	}
	else
	{
		if (fi.visual != FB_VISUAL_TRUECOLOR &&
		    fi.visual != FB_VISUAL_DIRECTCOLOR )
		{
			fprintf (stderr, "Error: need direct/truecolor framebuffer device.\n");
			exit(1);
		}
		else
		{
			printf ("Framebuffer resolution: %dx%d, %dbpp\n",
				vi.xres, vi.yres, vi.bits_per_pixel);

			fb = (unsigned long*) fi.smem_start;
			fblen = fi.smem_len;
			if (fblen < FBCOUNT) {
				printf ("Framebuffer is too small.\n");
				exit(1);
			}
			fb = mmap (fb, fblen,
				PROT_WRITE | PROT_READ,
				MAP_SHARED, fd, 0);
			if (fb == MAP_FAILED)
			{
				perror ("mmap");
				exit(1);
			}
		}
	}

	//-------------------
	// READ

	v = fb;
	loops = FBLOOPS_R;
	counter = FBCOUNT;

	printf ("Framebuffer memory sequential read ");
	fflush (stdout);

	t0 = time(NULL);
	while (loops--) {
		while (counter--)
			foo = *v++;

		v = fb;
		counter = FBCOUNT;
	}
	t = time(NULL);

	rate = FBCOUNT;
	rate *= FBLOOPS_R;
	rate /= (t-t0);
	rate /= 1000000;

	printf ("millions/sec = %g\n", rate);

	//-------------------
	// WRITE

	v = fb;
	loops = FBLOOPS_W;
	counter = FBCOUNT;

	printf ("Framebuffer memory sequential write ");
	fflush (stdout);

	t0 = time(NULL);
	while (loops--) {
		while (counter--)
			*v++ = 0x8080ffff;

		v = fb;
		counter = FBCOUNT;
	}
	t = time(NULL);

	rate = FBCOUNT;
	rate *= FBLOOPS_W;
	rate /= (t-t0);
	rate /= 1000000;

	printf ("millions/sec = %g\n", rate);
}


main()
{
	volatile unsigned long *chunk = (unsigned long*) malloc (4*MAINCOUNT);
	unsigned long register counter = 0;
	volatile unsigned long * v;
	register int loops;
	register unsigned long foo;
	time_t t0;

	printf ("This is bandwidth version %s\n", VERSION);
	printf ("Copyright (C) 2005 by Zack T Smith\n\n");

	l1_seq_read ();
	l1_seq_write ();

	l2_seq_read ();
	l2_seq_write ();

	main_seq_read ();
	main_seq_write ();

	fb_readwrite();

	return 0;
}
